from attrs import define, field
from typing import ClassVar, Dict, List, Optional, Type
from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.resource.base import AwsApiSpec, AwsResource, GraphBuilder
from resoto_plugin_aws.resource.iam import AwsIamRole
from resoto_plugin_aws.resource.kms import AwsKmsKey
from resoto_plugin_aws.resource.lambda_ import AwsLambdaFunction
from resotolib.baseresources import BaseUser, EdgeType, ModelReference
from resotolib.graph import Graph
from resotolib.json_bender import S, Bend, Bender, ForallBend
from resotolib.types import Json

service_name = "cognito-idp"


@define(eq=False, slots=False)
class AwsCognitoGroup(AwsResource):
    # collection of group resources happens in AwsCognitoUserPool.collect()
    kind: ClassVar[str] = "aws_cognito_group"
    reference_kinds: ClassVar[ModelReference] = {
        "predecessors": {"default": ["aws_iam_role"], "delete": ["aws_iam_role"]}
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("GroupName"),
        "name": S("GroupName"),
        "ctime": S("CreationDate"),
        "mtime": S("LastModifiedDate"),
        "user_pool_id": S("UserPoolId"),
        "description": S("Description"),
        "role_arn": S("RoleArn"),
        "precedence": S("Precedence"),
    }
    user_pool_id: Optional[str] = field(default=None)
    description: Optional[str] = field(default=None)
    role_arn: Optional[str] = field(default=None)
    precedence: Optional[int] = field(default=None)

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return [AwsApiSpec(service_name, "delete-group")]

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if self.role_arn:
            builder.dependant_node(self, reverse=True, delete_same_as_default=True, clazz=AwsIamRole, arn=self.role_arn)

    def delete_resource(self, client: AwsClient, graph: Graph) -> bool:
        client.call(
            aws_service=service_name, action="delete-group", result_name=None, GroupName=self.name, UserPoolId=self.id
        )
        return True


@define(eq=False, slots=False)
class AwsCognitoAttributeType:
    kind: ClassVar[str] = "aws_cognito_attribute_type"
    mapping: ClassVar[Dict[str, Bender]] = {"name": S("Name"), "value": S("Value")}
    name: Optional[str] = field(default=None)
    value: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCognitoMFAOptionType:
    kind: ClassVar[str] = "aws_cognito_mfa_option_type"
    mapping: ClassVar[Dict[str, Bender]] = {
        "delivery_medium": S("DeliveryMedium"),
        "attribute_name": S("AttributeName"),
    }
    delivery_medium: Optional[str] = field(default=None)
    attribute_name: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCognitoUser(AwsResource, BaseUser):
    # collection of user resources happens in AwsCognitoUserPool.collect()
    kind: ClassVar[str] = "aws_cognito_user"
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Username"),
        "name": S("Username"),
        "ctime": S("UserCreateDate"),
        "mtime": S("UserLastModifiedDate"),
        "user_attributes": S("Attributes", default=[]) >> ForallBend(AwsCognitoAttributeType.mapping),
        "enabled": S("Enabled"),
        "user_status": S("UserStatus"),
        "mfa_options": S("MFAOptions", default=[]) >> ForallBend(AwsCognitoMFAOptionType.mapping),
    }
    user_attributes: List[AwsCognitoAttributeType] = field(factory=list)
    enabled: Optional[bool] = field(default=None)
    user_status: Optional[str] = field(default=None)
    mfa_options: List[AwsCognitoMFAOptionType] = field(factory=list)


@define(eq=False, slots=False)
class AwsCognitoCustomSMSLambdaVersionConfigType:
    kind: ClassVar[str] = "aws_cognito_custom_sms_lambda_version_config_type"
    mapping: ClassVar[Dict[str, Bender]] = {"lambda_version": S("LambdaVersion"), "lambda_arn": S("LambdaArn")}
    lambda_version: Optional[str] = field(default=None)
    lambda_arn: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCognitoCustomEmailLambdaVersionConfigType:
    kind: ClassVar[str] = "aws_cognito_custom_email_lambda_version_config_type"
    mapping: ClassVar[Dict[str, Bender]] = {"lambda_version": S("LambdaVersion"), "lambda_arn": S("LambdaArn")}
    lambda_version: Optional[str] = field(default=None)
    lambda_arn: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCognitoLambdaConfigType:
    kind: ClassVar[str] = "aws_cognito_lambda_config_type"
    mapping: ClassVar[Dict[str, Bender]] = {
        "pre_sign_up": S("PreSignUp"),
        "custom_message": S("CustomMessage"),
        "post_confirmation": S("PostConfirmation"),
        "pre_authentication": S("PreAuthentication"),
        "post_authentication": S("PostAuthentication"),
        "define_auth_challenge": S("DefineAuthChallenge"),
        "create_auth_challenge": S("CreateAuthChallenge"),
        "verify_auth_challenge_response": S("VerifyAuthChallengeResponse"),
        "pre_token_generation": S("PreTokenGeneration"),
        "user_migration": S("UserMigration"),
        "custom_sms_sender": S("CustomSMSSender") >> Bend(AwsCognitoCustomSMSLambdaVersionConfigType.mapping),
        "custom_email_sender": S("CustomEmailSender") >> Bend(AwsCognitoCustomEmailLambdaVersionConfigType.mapping),
        "kms_key_id": S("KMSKeyID"),
    }
    pre_sign_up: Optional[str] = field(default=None)
    custom_message: Optional[str] = field(default=None)
    post_confirmation: Optional[str] = field(default=None)
    pre_authentication: Optional[str] = field(default=None)
    post_authentication: Optional[str] = field(default=None)
    define_auth_challenge: Optional[str] = field(default=None)
    create_auth_challenge: Optional[str] = field(default=None)
    verify_auth_challenge_response: Optional[str] = field(default=None)
    pre_token_generation: Optional[str] = field(default=None)
    user_migration: Optional[str] = field(default=None)
    custom_sms_sender: Optional[AwsCognitoCustomSMSLambdaVersionConfigType] = field(default=None)
    custom_email_sender: Optional[AwsCognitoCustomEmailLambdaVersionConfigType] = field(default=None)
    kms_key_id: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCognitoUserPool(AwsResource):
    kind: ClassVar[str] = "aws_cognito_user_pool"
    # this call requires the MaxResult parameter, 60 is the maximum valid input
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec(service_name, "list-user-pools", "UserPools", {"MaxResults": 60})
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {"default": ["aws_cognito_user", "aws_cognito_group", "aws_lambda_function", "aws_kms_key"]},
        "predecessors": {"delete": ["aws_lambda_function", "aws_kms_key"]},
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Id"),
        "name": S("Name"),
        "lambda_config": S("LambdaConfig") >> Bend(AwsCognitoLambdaConfigType.mapping),
        "status": S("Status"),
        "mtime": S("LastModifiedDate"),
        "ctime": S("CreationDate"),
    }
    lambda_config: Optional[AwsCognitoLambdaConfigType] = field(default=None)
    status: Optional[str] = field(default=None)

    @classmethod
    def called_collect_apis(cls) -> List[AwsApiSpec]:
        return [
            cls.api_spec,
            AwsApiSpec(service_name, "list-tags-for-resource"),
            AwsApiSpec(service_name, "list-users"),
            AwsApiSpec(service_name, "list-groups"),
        ]

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return [
            AwsApiSpec(service_name, "tag-resource"),
            AwsApiSpec(service_name, "untag-resource"),
            AwsApiSpec(service_name, "delete-user-pool"),
        ]

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        def add_tags(pool: AwsCognitoUserPool) -> None:
            tags = builder.client.get(service_name, "list-tags-for-resource", "Tags", ResourceArn=pool.arn)
            if tags:
                pool.tags = tags

        for pool in json:
            pool_instance = cls.from_api(pool)
            pool_instance.set_arn(builder=builder, resource=f"userpool/{pool_instance.id}")
            builder.add_node(pool_instance, pool)
            builder.submit_work(service_name, add_tags, pool_instance)
            for user in builder.client.list(service_name, "list-users", "Users", UserPoolId=pool_instance.id):
                user_instance = AwsCognitoUser.from_api(user)
                builder.add_node(user_instance, user)
                builder.add_edge(from_node=pool_instance, edge_type=EdgeType.default, node=user_instance)
            for group in builder.client.list(service_name, "list-groups", "Groups", UserPoolId=pool_instance.id):
                group_instance = AwsCognitoGroup.from_api(group)
                builder.add_node(group_instance, group)
                builder.add_edge(from_node=pool_instance, edge_type=EdgeType.default, node=group_instance)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if self.lambda_config:
            if self.lambda_config.custom_sms_sender:
                builder.dependant_node(
                    self,
                    clazz=AwsLambdaFunction,
                    arn=self.lambda_config.custom_sms_sender.lambda_arn,
                )
            if self.lambda_config.custom_email_sender:
                builder.dependant_node(
                    self,
                    clazz=AwsLambdaFunction,
                    arn=self.lambda_config.custom_email_sender.lambda_arn,
                )
            if self.lambda_config.kms_key_id:
                builder.dependant_node(
                    self,
                    clazz=AwsKmsKey,
                    id=AwsKmsKey.normalise_id(self.lambda_config.kms_key_id),
                )

    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        client.call(
            aws_service=service_name, action="tag-resource", result_name=None, ResourceArn=self.arn, Tags={key: value}
        )
        return True

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        client.call(
            aws_service=service_name, action="untag-resource", result_name=None, ResourceArn=self.arn, TagKeys=[key]
        )
        return True

    def delete_resource(self, client: AwsClient, graph: Graph) -> bool:
        client.call(aws_service=service_name, action="delete-user-pool", result_name=None, UserPoolId=self.id)
        return True


resources: List[Type[AwsResource]] = [AwsCognitoUserPool, AwsCognitoUser, AwsCognitoGroup]



from datetime import datetime
from typing import ClassVar, Dict, List, Optional, Type
from attrs import define, field
from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.resource.base import AwsApiSpec, AwsResource, GraphBuilder
from resoto_plugin_aws.resource.kinesis import AwsKinesisStream
from resoto_plugin_aws.resource.kms import AwsKmsKey
from resoto_plugin_aws.utils import ToDict
from resotolib.baseresources import ModelReference
from resotolib.graph import Graph
from resotolib.json_bender import S, Bend, Bender, ForallBend, bend
from resotolib.types import Json

service_name = "dynamodb"


# noinspection PyUnresolvedReferences
class DynamoDbTaggable:
    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        if isinstance(self, AwsResource):
            client.call(
                aws_service=service_name,
                action="tag-resource",
                result_name=None,
                ResourceArn=self.arn,
                Tags=[{"Key": key, "Value": value}],
            )
            return True
        return False

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        if isinstance(self, AwsResource):
            client.call(
                aws_service=service_name,
                action="untag-resource",
                result_name=None,
                ResourceArn=self.arn,
                TagKeys=[key],
            )
            return True
        return False

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return [AwsApiSpec(service_name, "tag-resource"), AwsApiSpec(service_name, "untag-resource")]


@define(eq=False, slots=False)
class AwsDynamoDbAttributeDefinition:
    kind: ClassVar[str] = "aws_dynamo_db_attribute_definition"
    mapping: ClassVar[Dict[str, Bender]] = {"attribute_name": S("AttributeName"), "attribute_type": S("AttributeType")}
    attribute_name: Optional[str] = field(default=None)
    attribute_type: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsDynamoDbKeySchemaElement:
    kind: ClassVar[str] = "aws_dynamo_db_key_schema_element"
    mapping: ClassVar[Dict[str, Bender]] = {"attribute_name": S("AttributeName"), "key_type": S("KeyType")}
    attribute_name: Optional[str] = field(default=None)
    key_type: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsDynamoDbProvisionedThroughputDescription:
    kind: ClassVar[str] = "aws_dynamo_db_provisioned_throughput_description"
    mapping: ClassVar[Dict[str, Bender]] = {
        "last_increase_date_time": S("LastIncreaseDateTime"),
        "last_decrease_date_time": S("LastDecreaseDateTime"),
        "number_of_decreases_today": S("NumberOfDecreasesToday"),
        "read_capacity_units": S("ReadCapacityUnits"),
        "write_capacity_units": S("WriteCapacityUnits"),
    }
    last_increase_date_time: Optional[datetime] = field(default=None)
    last_decrease_date_time: Optional[datetime] = field(default=None)
    number_of_decreases_today: Optional[int] = field(default=None)
    read_capacity_units: Optional[int] = field(default=None)
    write_capacity_units: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class AwsDynamoDbBillingModeSummary:
    kind: ClassVar[str] = "aws_dynamo_db_billing_mode_summary"
    mapping: ClassVar[Dict[str, Bender]] = {
        "billing_mode": S("BillingMode"),
        "last_update_to_pay_per_request_date_time": S("LastUpdateToPayPerRequestDateTime"),
    }
    billing_mode: Optional[str] = field(default=None)
    last_update_to_pay_per_request_date_time: Optional[datetime] = field(default=None)


@define(eq=False, slots=False)
class AwsDynamoDbProjection:
    kind: ClassVar[str] = "aws_dynamo_db_projection"
    mapping: ClassVar[Dict[str, Bender]] = {
        "projection_type": S("ProjectionType"),
        "non_key_attributes": S("NonKeyAttributes", default=[]),
    }
    projection_type: Optional[str] = field(default=None)
    non_key_attributes: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsDynamoDbLocalSecondaryIndexDescription:
    kind: ClassVar[str] = "aws_dynamo_db_local_secondary_index_description"
    mapping: ClassVar[Dict[str, Bender]] = {
        "index_name": S("IndexName"),
        "key_schema": S("KeySchema", default=[]) >> ForallBend(AwsDynamoDbKeySchemaElement.mapping),
        "projection": S("Projection") >> Bend(AwsDynamoDbProjection.mapping),
        "index_size_bytes": S("IndexSizeBytes"),
        "item_count": S("ItemCount"),
        "index_arn": S("IndexArn"),
    }
    index_name: Optional[str] = field(default=None)
    key_schema: List[AwsDynamoDbKeySchemaElement] = field(factory=list)
    projection: Optional[AwsDynamoDbProjection] = field(default=None)
    index_size_bytes: Optional[int] = field(default=None)
    item_count: Optional[int] = field(default=None)
    index_arn: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsDynamoDbGlobalSecondaryIndexDescription:
    kind: ClassVar[str] = "aws_dynamo_db_global_secondary_index_description"
    mapping: ClassVar[Dict[str, Bender]] = {
        "index_name": S("IndexName"),
        "key_schema": S("KeySchema", default=[]) >> ForallBend(AwsDynamoDbKeySchemaElement.mapping),
        "projection": S("Projection") >> Bend(AwsDynamoDbProjection.mapping),
        "index_status": S("IndexStatus"),
        "backfilling": S("Backfilling"),
        "provisioned_throughput": S("ProvisionedThroughput")
        >> Bend(AwsDynamoDbProvisionedThroughputDescription.mapping),
        "index_size_bytes": S("IndexSizeBytes"),
        "item_count": S("ItemCount"),
        "index_arn": S("IndexArn"),
    }
    index_name: Optional[str] = field(default=None)
    key_schema: List[AwsDynamoDbKeySchemaElement] = field(factory=list)
    projection: Optional[AwsDynamoDbProjection] = field(default=None)
    index_status: Optional[str] = field(default=None)
    backfilling: Optional[bool] = field(default=None)
    provisioned_throughput: Optional[AwsDynamoDbProvisionedThroughputDescription] = field(default=None)
    index_size_bytes: Optional[int] = field(default=None)
    item_count: Optional[int] = field(default=None)
    index_arn: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsDynamoDbStreamSpecification:
    kind: ClassVar[str] = "aws_dynamo_db_stream_specification"
    mapping: ClassVar[Dict[str, Bender]] = {
        "stream_enabled": S("StreamEnabled"),
        "stream_view_type": S("StreamViewType"),
    }
    stream_enabled: Optional[bool] = field(default=None)
    stream_view_type: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsDynamoDbReplicaGlobalSecondaryIndexDescription:
    kind: ClassVar[str] = "aws_dynamo_db_replica_global_secondary_index_description"
    mapping: ClassVar[Dict[str, Bender]] = {
        "index_name": S("IndexName"),
        "provisioned_throughput_override": S("ProvisionedThroughputOverride", "ReadCapacityUnits"),
    }
    index_name: Optional[str] = field(default=None)
    provisioned_throughput_override: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class AwsDynamoDbTableClassSummary:
    kind: ClassVar[str] = "aws_dynamo_db_table_class_summary"
    mapping: ClassVar[Dict[str, Bender]] = {
        "table_class": S("TableClass"),
        "last_update_date_time": S("LastUpdateDateTime"),
    }
    table_class: Optional[str] = field(default=None)
    last_update_date_time: Optional[datetime] = field(default=None)


@define(eq=False, slots=False)
class AwsDynamoDbReplicaDescription:
    kind: ClassVar[str] = "aws_dynamo_db_replica_description"
    mapping: ClassVar[Dict[str, Bender]] = {
        "region_name": S("RegionName"),
        "replica_status": S("ReplicaStatus"),
        "replica_status_description": S("ReplicaStatusDescription"),
        "replica_status_percent_progress": S("ReplicaStatusPercentProgress"),
        "kms_master_key_id": S("KMSMasterKeyId"),
        "provisioned_throughput_override": S("ProvisionedThroughputOverride", "ReadCapacityUnits"),
        "global_secondary_indexes": S("GlobalSecondaryIndexes", default=[])
        >> ForallBend(AwsDynamoDbReplicaGlobalSecondaryIndexDescription.mapping),
        "replica_inaccessible_date_time": S("ReplicaInaccessibleDateTime"),
        "replica_table_class_summary": S("ReplicaTableClassSummary") >> Bend(AwsDynamoDbTableClassSummary.mapping),
    }
    region_name: Optional[str] = field(default=None)
    replica_status: Optional[str] = field(default=None)
    replica_status_description: Optional[str] = field(default=None)
    replica_status_percent_progress: Optional[str] = field(default=None)
    kms_master_key_id: Optional[str] = field(default=None)
    provisioned_throughput_override: Optional[int] = field(default=None)
    global_secondary_indexes: List[AwsDynamoDbReplicaGlobalSecondaryIndexDescription] = field(factory=list)
    replica_inaccessible_date_time: Optional[datetime] = field(default=None)
    replica_table_class_summary: Optional[AwsDynamoDbTableClassSummary] = field(default=None)


@define(eq=False, slots=False)
class AwsDynamoDbRestoreSummary:
    kind: ClassVar[str] = "aws_dynamo_db_restore_summary"
    mapping: ClassVar[Dict[str, Bender]] = {
        "source_backup_arn": S("SourceBackupArn"),
        "source_table_arn": S("SourceTableArn"),
        "restore_date_time": S("RestoreDateTime"),
        "restore_in_progress": S("RestoreInProgress"),
    }
    source_backup_arn: Optional[str] = field(default=None)
    source_table_arn: Optional[str] = field(default=None)
    restore_date_time: Optional[datetime] = field(default=None)
    restore_in_progress: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class AwsDynamoDbSSEDescription:
    kind: ClassVar[str] = "aws_dynamo_db_sse_description"
    mapping: ClassVar[Dict[str, Bender]] = {
        "status": S("Status"),
        "sse_type": S("SSEType"),
        "kms_master_key_arn": S("KMSMasterKeyArn"),
        "inaccessible_encryption_date_time": S("InaccessibleEncryptionDateTime"),
    }
    status: Optional[str] = field(default=None)
    sse_type: Optional[str] = field(default=None)
    kms_master_key_arn: Optional[str] = field(default=None)
    inaccessible_encryption_date_time: Optional[datetime] = field(default=None)


@define(eq=False, slots=False)
class AwsDynamoDbArchivalSummary:
    kind: ClassVar[str] = "aws_dynamo_db_archival_summary"
    mapping: ClassVar[Dict[str, Bender]] = {
        "archival_date_time": S("ArchivalDateTime"),
        "archival_reason": S("ArchivalReason"),
        "archival_backup_arn": S("ArchivalBackupArn"),
    }
    archival_date_time: Optional[datetime] = field(default=None)
    archival_reason: Optional[str] = field(default=None)
    archival_backup_arn: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsDynamoDbTable(DynamoDbTaggable, AwsResource):
    kind: ClassVar[str] = "aws_dynamo_db_table"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec(service_name, "list-tables", "TableNames")
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {"default": ["aws_kinesis_stream", "aws_kms_key"]},
        "predecessors": {"delete": ["aws_kinesis_stream", "aws_kms_key"]},
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("TableId"),
        "name": S("TableName"),
        "ctime": S("CreationDateTime"),
        "arn": S("TableArn"),
        "dynamodb_attribute_definitions": S("AttributeDefinitions", default=[])
        >> ForallBend(AwsDynamoDbAttributeDefinition.mapping),
        "dynamodb_key_schema": S("KeySchema", default=[]) >> ForallBend(AwsDynamoDbKeySchemaElement.mapping),
        "dynamodb_table_status": S("TableStatus"),
        "dynamodb_provisioned_throughput": S("ProvisionedThroughput")
        >> Bend(AwsDynamoDbProvisionedThroughputDescription.mapping),
        "dynamodb_table_size_bytes": S("TableSizeBytes"),
        "dynamodb_item_count": S("ItemCount"),
        "dynamodb_billing_mode_summary": S("BillingModeSummary") >> Bend(AwsDynamoDbBillingModeSummary.mapping),
        "dynamodb_local_secondary_indexes": S("LocalSecondaryIndexes", default=[])
        >> ForallBend(AwsDynamoDbLocalSecondaryIndexDescription.mapping),
        "dynamodb_global_secondary_indexes": S("GlobalSecondaryIndexes", default=[])
        >> ForallBend(AwsDynamoDbGlobalSecondaryIndexDescription.mapping),
        "dynamodb_stream_specification": S("StreamSpecification") >> Bend(AwsDynamoDbStreamSpecification.mapping),
        "dynamodb_latest_stream_label": S("LatestStreamLabel"),
        "dynamodb_latest_stream_arn": S("LatestStreamArn"),
        "dynamodb_global_table_version": S("GlobalTableVersion"),
        "dynamodb_replicas": S("Replicas", default=[]) >> ForallBend(AwsDynamoDbReplicaDescription.mapping),
        "dynamodb_restore_summary": S("RestoreSummary") >> Bend(AwsDynamoDbRestoreSummary.mapping),
        "dynamodb_sse_description": S("SSEDescription") >> Bend(AwsDynamoDbSSEDescription.mapping),
        "dynamodb_archival_summary": S("ArchivalSummary") >> Bend(AwsDynamoDbArchivalSummary.mapping),
        "dynamodb_table_class_summary": S("TableClassSummary") >> Bend(AwsDynamoDbTableClassSummary.mapping),
    }
    arn: Optional[str] = field(default=None)
    dynamodb_attribute_definitions: List[AwsDynamoDbAttributeDefinition] = field(factory=list)
    dynamodb_key_schema: List[AwsDynamoDbKeySchemaElement] = field(factory=list)
    dynamodb_table_status: Optional[str] = field(default=None)
    dynamodb_provisioned_throughput: Optional[AwsDynamoDbProvisionedThroughputDescription] = field(default=None)
    dynamodb_table_size_bytes: Optional[int] = field(default=None)
    dynamodb_item_count: Optional[int] = field(default=None)
    dynamodb_billing_mode_summary: Optional[AwsDynamoDbBillingModeSummary] = field(default=None)
    dynamodb_local_secondary_indexes: List[AwsDynamoDbLocalSecondaryIndexDescription] = field(factory=list)
    dynamodb_global_secondary_indexes: List[AwsDynamoDbGlobalSecondaryIndexDescription] = field(factory=list)
    dynamodb_stream_specification: Optional[AwsDynamoDbStreamSpecification] = field(default=None)
    dynamodb_latest_stream_label: Optional[str] = field(default=None)
    dynamodb_latest_stream_arn: Optional[str] = field(default=None)
    dynamodb_global_table_version: Optional[str] = field(default=None)
    dynamodb_replicas: List[AwsDynamoDbReplicaDescription] = field(factory=list)
    dynamodb_restore_summary: Optional[AwsDynamoDbRestoreSummary] = field(default=None)
    dynamodb_sse_description: Optional[AwsDynamoDbSSEDescription] = field(default=None)
    dynamodb_archival_summary: Optional[AwsDynamoDbArchivalSummary] = field(default=None)
    dynamodb_table_class_summary: Optional[AwsDynamoDbTableClassSummary] = field(default=None)

    @classmethod
    def called_collect_apis(cls) -> List[AwsApiSpec]:
        return [
            cls.api_spec,
            AwsApiSpec(service_name, "describe-table"),
            AwsApiSpec(service_name, "list-tags-of-resource"),
        ]

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        def add_instance(table: str) -> None:
            table_description = builder.client.get(service_name, "describe-table", "Table", TableName=table)
            if table_description is not None:
                instance = cls.from_api(table_description)
                builder.add_node(instance, table_description)
                builder.submit_work(service_name, add_tags, instance)

        def add_tags(table: AwsDynamoDbTable) -> None:
            tags = builder.client.list(service_name, "list-tags-of-resource", "Tags", ResourceArn=table.arn)
            if tags:
                table.tags = bend(ToDict(), tags)

        for js in json:
            if isinstance(js, str):
                add_instance(js)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if self.dynamodb_latest_stream_arn:
            builder.dependant_node(
                self,
                clazz=AwsKinesisStream,
                arn=self.dynamodb_latest_stream_arn,
            )
        for replica in self.dynamodb_replicas:
            if replica.kms_master_key_id:
                builder.dependant_node(
                    self,
                    clazz=AwsKmsKey,
                    id=replica.kms_master_key_id,
                )
        if self.dynamodb_sse_description and self.dynamodb_sse_description.kms_master_key_arn:
            builder.dependant_node(
                self,
                clazz=AwsKmsKey,
                arn=self.dynamodb_sse_description.kms_master_key_arn,
            )

    def delete_resource(self, client: AwsClient, graph: Graph) -> bool:
        client.call(aws_service=self.api_spec.service, action="delete-table", result_name=None, TableName=self.name)
        return True

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return super().called_mutator_apis() + [AwsApiSpec(service_name, "delete-table")]


@define(eq=False, slots=False)
class AwsDynamoDbGlobalTable(DynamoDbTaggable, AwsResource):
    kind: ClassVar[str] = "aws_dynamo_db_global_table"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec(service_name, "list-global-tables", "GlobalTables")
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {"default": ["aws_kms_key"]},
        "predecessors": {"delete": ["aws_kms_key"]},
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("GlobalTableName"),
        "name": S("GlobalTableName"),
        "ctime": S("CreationDateTime"),
        "arn": S("GlobalTableArn"),
        "dynamodb_replication_group": S("ReplicationGroup", default=[])
        >> ForallBend(AwsDynamoDbReplicaDescription.mapping),
        "dynamodb_global_table_status": S("GlobalTableStatus"),
    }
    arn: Optional[str] = field(default=None)
    dynamodb_replication_group: List[AwsDynamoDbReplicaDescription] = field(factory=list)
    dynamodb_global_table_status: Optional[str] = field(default=None)

    @classmethod
    def called_collect_apis(cls) -> List[AwsApiSpec]:
        return [
            cls.api_spec,
            AwsApiSpec(service_name, "describe-global-table"),
            AwsApiSpec(service_name, "list-tags-of-resource"),
        ]

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        def add_instance(table: Dict[str, str]) -> None:
            table_description = builder.client.get(
                service_name,
                "describe-global-table",
                "GlobalTableDescription",
                GlobalTableName=table["GlobalTableName"],
            )
            if table_description:
                instance = cls.from_api(table_description)
                builder.add_node(instance, table_description)
                builder.submit_work(service_name, add_tags, instance)

        def add_tags(table: AwsDynamoDbGlobalTable) -> None:
            tags = builder.client.list(service_name, "list-tags-of-resource", "Tags", ResourceArn=table.arn)
            if tags:
                table.tags = bend(ToDict(), tags)

        for js in json:
            add_instance(js)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if self.dynamodb_replication_group is not []:
            for replica in self.dynamodb_replication_group:
                if replica.kms_master_key_id:
                    builder.dependant_node(
                        self,
                        clazz=AwsKmsKey,
                        id=replica.kms_master_key_id,
                    )

    def delete_resource(self, client: AwsClient, graph: Graph) -> bool:
        client.call(aws_service=self.api_spec.service, action="delete-table", result_name=None, TableName=self.name)
        return True

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return super().called_mutator_apis() + [AwsApiSpec(service_name, "delete-table")]


global_resources: List[Type[AwsResource]] = [AwsDynamoDbGlobalTable]
resources: List[Type[AwsResource]] = [AwsDynamoDbTable]



from contextlib import suppress
from json import loads as json_loads
from typing import ClassVar, Dict, List, Type, Optional, cast, Any

from attr import field
from attrs import define

from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.resource.base import AwsResource, AwsApiSpec, GraphBuilder
from resoto_plugin_aws.utils import tags_as_dict
from resotolib.baseresources import BaseBucket, PhantomBaseResource, ModelReference
from resotolib.graph import Graph
from resotolib.json import from_json, is_empty
from resotolib.json_bender import Bender, S, bend, Bend, ForallBend
from resotolib.types import Json

service_name = "s3"


@define(eq=False, slots=False)
class AwsS3ServerSideEncryptionRule:
    kind: ClassVar[str] = "aws_s3_server_side_encryption_rule"
    mapping: ClassVar[Dict[str, Bender]] = {
        "sse_algorithm": S("ApplyServerSideEncryptionByDefault", "SSEAlgorithm"),
        "kms_master_key_id": S("ApplyServerSideEncryptionByDefault", "KMSMasterKeyID"),
        "bucket_key_enabled": S("BucketKeyEnabled"),
    }
    sse_algorithm: Optional[str] = field(default=None)
    kms_master_key_id: Optional[str] = field(default=None)
    bucket_key_enabled: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class AwsS3PublicAccessBlockConfiguration:
    kind: ClassVar[str] = "aws_s3_public_access_block_configuration"
    mapping: ClassVar[Dict[str, Bender]] = {
        "block_public_acls": S("BlockPublicAcls"),
        "ignore_public_acls": S("IgnorePublicAcls"),
        "block_public_policy": S("BlockPublicPolicy"),
        "restrict_public_buckets": S("RestrictPublicBuckets"),
    }
    block_public_acls: Optional[bool] = field(default=False)
    ignore_public_acls: Optional[bool] = field(default=False)
    block_public_policy: Optional[bool] = field(default=False)
    restrict_public_buckets: Optional[bool] = field(default=False)


@define(eq=False, slots=False)
class AwsS3Owner:
    kind: ClassVar[str] = "aws_s3_owner"
    mapping: ClassVar[Dict[str, Bender]] = {"display_name": S("DisplayName"), "id": S("ID")}
    display_name: Optional[str] = field(default=None)
    id: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsS3Grantee:
    kind: ClassVar[str] = "aws_s3_grantee"
    mapping: ClassVar[Dict[str, Bender]] = {
        "display_name": S("DisplayName"),
        "email_address": S("EmailAddress"),
        "id": S("ID"),
        "type": S("Type"),
        "uri": S("URI"),
    }
    display_name: Optional[str] = field(default=None)
    email_address: Optional[str] = field(default=None)
    id: Optional[str] = field(default=None)
    type: Optional[str] = field(default=None)
    uri: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsS3Grant:
    kind: ClassVar[str] = "aws_s3_grant"
    mapping: ClassVar[Dict[str, Bender]] = {
        "grantee": S("Grantee") >> Bend(AwsS3Grantee.mapping),
        "permission": S("Permission"),
    }
    grantee: Optional[AwsS3Grantee] = field(default=None)
    permission: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsS3BucketAcl:
    kind: ClassVar[str] = "aws_s3_bucket_acl"
    mapping: ClassVar[Dict[str, Bender]] = {
        "owner": S("Owner") >> Bend(AwsS3Owner.mapping),
        "grants": S("Grants", default=[]) >> ForallBend(AwsS3Grant.mapping),
    }
    owner: Optional[AwsS3Owner] = field(default=None)
    grants: List[AwsS3Grant] = field(factory=list)


@define(eq=False, slots=False)
class AwsS3TargetGrant:
    kind: ClassVar[str] = "aws_s3_target_grant"
    mapping: ClassVar[Dict[str, Bender]] = {
        "grantee": S("Grantee") >> Bend(AwsS3Grantee.mapping),
        "permission": S("Permission"),
    }
    grantee: Optional[AwsS3Grantee] = field(default=None)
    permission: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsS3Logging:
    kind: ClassVar[str] = "aws_s3_logging"
    mapping: ClassVar[Dict[str, Bender]] = {
        "target_bucket": S("TargetBucket"),
        "target_grants": S("TargetGrants") >> ForallBend(AwsS3TargetGrant.mapping),
        "target_prefix": S("TargetPrefix"),
    }
    target_bucket: Optional[str] = field(default=None)
    target_grants: Optional[List[AwsS3TargetGrant]] = field(default=None)
    target_prefix: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsS3Bucket(AwsResource, BaseBucket):
    kind: ClassVar[str] = "aws_s3_bucket"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec(
        service_name, "list-buckets", "Buckets", override_iam_permission="s3:ListAllMyBuckets"
    )
    mapping: ClassVar[Dict[str, Bender]] = {"id": S("Name"), "name": S("Name"), "ctime": S("CreationDate")}
    bucket_encryption_rules: Optional[List[AwsS3ServerSideEncryptionRule]] = field(default=None)
    bucket_policy: Optional[Json] = field(default=None)
    bucket_versioning: Optional[bool] = field(default=None)
    bucket_mfa_delete: Optional[bool] = field(default=None)
    bucket_public_access_block_configuration: Optional[AwsS3PublicAccessBlockConfiguration] = field(default=None)
    bucket_acl: Optional[AwsS3BucketAcl] = field(default=None)
    bucket_logging: Optional[AwsS3Logging] = field(default=None)

    @classmethod
    def called_collect_apis(cls) -> List[AwsApiSpec]:
        return [
            cls.api_spec,
            AwsApiSpec(service_name, "get-bucket-tagging"),
            AwsApiSpec(service_name, "get-bucket-encryption", override_iam_permission="s3:GetEncryptionConfiguration"),
            AwsApiSpec(service_name, "get-bucket-policy"),
            AwsApiSpec(service_name, "get-bucket-versioning"),
            AwsApiSpec(
                service_name, "get-public-access-block", override_iam_permission="s3:GetAccountPublicAccessBlock"
            ),
            AwsApiSpec(service_name, "get-bucket-acl"),
            AwsApiSpec(service_name, "get-bucket-logging"),
        ]

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        def add_tags(bucket: AwsS3Bucket) -> None:
            tags = bucket._get_tags(builder.client)
            if tags:
                bucket.tags = cast(Dict[str, Optional[str]], tags)

        def add_bucket_encryption(bck: AwsS3Bucket) -> None:
            with suppress(Exception):
                bck.bucket_encryption_rules = []
                for raw in builder.client.list(
                    service_name,
                    "get-bucket-encryption",
                    "ServerSideEncryptionConfiguration.Rules",
                    Bucket=bck.name,
                    expected_errors=["ServerSideEncryptionConfigurationNotFoundError", "NoSuchBucket"],
                ):
                    mapped = bend(AwsS3ServerSideEncryptionRule.mapping, raw)
                    bck.bucket_encryption_rules.append(from_json(mapped, AwsS3ServerSideEncryptionRule))

        def add_bucket_policy(bck: AwsS3Bucket) -> None:
            with suppress(Exception):
                if raw_policy := builder.client.get(
                    service_name,
                    "get-bucket-policy",
                    "Policy",
                    Bucket=bck.name,
                    expected_errors=["NoSuchBucketPolicy", "NoSuchBucket"],
                ):
                    bck.bucket_policy = json_loads(raw_policy)  # type: ignore # this is a string

        def add_bucket_versioning(bck: AwsS3Bucket) -> None:
            with suppress(Exception):
                if raw_versioning := builder.client.get(
                    service_name, "get-bucket-versioning", None, Bucket=bck.name, expected_errors=["NoSuchBucket"]
                ):
                    bck.bucket_versioning = raw_versioning.get("Status") == "Enabled"
                    bck.bucket_mfa_delete = raw_versioning.get("MFADelete") == "Enabled"
                else:
                    bck.bucket_versioning = False
                    bck.bucket_mfa_delete = False

        def add_public_access(bck: AwsS3Bucket) -> None:
            with suppress(Exception):
                if raw_access := builder.client.get(
                    service_name,
                    "get-public-access-block",
                    "PublicAccessBlockConfiguration",
                    Bucket=bck.name,
                    expected_errors=["NoSuchPublicAccessBlockConfiguration", "NoSuchBucket"],
                ):
                    mapped = bend(AwsS3PublicAccessBlockConfiguration.mapping, raw_access)
                    bck.bucket_public_access_block_configuration = from_json(
                        mapped, AwsS3PublicAccessBlockConfiguration
                    )

        def add_acls(bck: AwsS3Bucket) -> None:
            with suppress(Exception):
                if raw := builder.client.get(
                    service_name, "get-bucket-acl", Bucket=bck.name, expected_errors=["NoSuchBucket"]
                ):
                    mapped = bend(AwsS3BucketAcl.mapping, raw)
                    bck.bucket_acl = from_json(mapped, AwsS3BucketAcl)

        def add_bucket_logging(bck: AwsS3Bucket) -> None:
            with suppress(Exception):
                if raw := builder.client.get(
                    service_name,
                    "get-bucket-logging",
                    "LoggingEnabled",
                    Bucket=bck.name,
                    expected_errors=["NoSuchBucket"],
                ):
                    mapped = bend(AwsS3Logging.mapping, raw)
                    # do not set, if no property is set
                    if not is_empty(mapped):
                        bck.bucket_logging = from_json(mapped, AwsS3Logging)

        for js in json:
            bucket = cls.from_api(js)
            bucket.set_arn(builder=builder, region="", account="", resource=bucket.safe_name)
            builder.add_node(bucket, js)
            builder.submit_work(service_name, add_tags, bucket)
            builder.submit_work(service_name, add_bucket_encryption, bucket)
            builder.submit_work(service_name, add_bucket_policy, bucket)
            builder.submit_work(service_name, add_bucket_versioning, bucket)
            builder.submit_work(service_name, add_public_access, bucket)
            builder.submit_work(service_name, add_acls, bucket)
            builder.submit_work(service_name, add_bucket_logging, bucket)

    def _set_tags(self, client: AwsClient, tags: Dict[str, str]) -> bool:
        tag_set = [{"Key": k, "Value": v} for k, v in tags.items()]
        client.call(
            aws_service=service_name,
            action="put-bucket-tagging",
            result_name=None,
            Bucket=self.name,
            Tagging={"TagSet": tag_set},
        )
        return True

    def _get_tags(self, client: AwsClient) -> Dict[str, str]:
        """Fetch the S3 buckets tags from the AWS API."""
        tag_list = client.list(
            aws_service=service_name,
            action="get-bucket-tagging",
            result_name="TagSet",
            expected_errors=["NoSuchTagSet", "NoSuchBucket"],
            Bucket=self.name,
        )
        return tags_as_dict(tag_list)  # type: ignore

    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        tags = self._get_tags(client)
        tags[key] = value
        return self._set_tags(client, tags)

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        tags = self._get_tags(client)
        if key in tags:
            del tags[key]
        else:
            raise KeyError(key)
        return self._set_tags(client, tags)

    def delete_resource(self, client: AwsClient, graph: Graph) -> bool:
        def delete_bucket_content(s3: Any) -> bool:
            bucket = s3.Bucket(self.name)
            bucket.objects.delete()
            bucket.delete()
            return True

        return client.with_resource(service_name, delete_bucket_content) or False

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return [
            AwsApiSpec(service_name, "put-bucket-tagging"),
            AwsApiSpec(service_name, "delete-object"),
            AwsApiSpec(service_name, "delete-bucket"),
        ]

    @staticmethod
    def name_from_path(path_or_uri: str) -> str:
        # https://docs.aws.amazon.com/AmazonS3/latest/userguide/access-bucket-intro.html
        # Accessing a bucket using S3://
        if path_or_uri.startswith("s3://") or path_or_uri.startswith("S3://"):
            return path_or_uri.split("/")[2]
        # Path-style access
        if path_or_uri.startswith("https://s3"):
            bucket_and_key = path_or_uri.split("amazonaws.com/")[-1]
            return bucket_and_key.split("/")[0]
        # Virtual-hosted–style access
        if path_or_uri.startswith("https://"):
            bucket_and_key = path_or_uri.split("//")[-1]
            return bucket_and_key.split(".")[0]
        return path_or_uri


@define(eq=False)
class AwsS3AccountSettings(AwsResource, PhantomBaseResource):
    """
    This resource is fetched once for every account.
    """

    kind: ClassVar[str] = "aws_s3_account_settings"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {"default": ["aws_account"]},
    }
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec(
        "s3control",
        "get-public-access-block",
        override_iam_permission="s3:GetAccountPublicAccessBlock",
    )

    bucket_public_access_block_configuration: Optional[AwsS3PublicAccessBlockConfiguration] = field(default=None)

    @classmethod
    def collect_resources(cls: Type[AwsResource], builder: GraphBuilder) -> None:
        node = AwsS3AccountSettings(
            id=builder.account.id,
            name=builder.account.name,
            ctime=builder.account.ctime,
            bucket_public_access_block_configuration=AwsS3PublicAccessBlockConfiguration(),
        )
        if raw := builder.client.get(
            "s3control",
            "get-public-access-block",
            "PublicAccessBlockConfiguration",
            AccountId=builder.account.id,
            expected_errors=["NoSuchPublicAccessBlockConfiguration"],
        ):
            mapped = bend(AwsS3PublicAccessBlockConfiguration.mapping, raw)
            node.bucket_public_access_block_configuration = from_json(mapped, AwsS3PublicAccessBlockConfiguration)
        builder.add_node(node)
        builder.add_edge(builder.account, node=node)


resources: List[Type[AwsResource]] = [AwsS3AccountSettings, AwsS3Bucket]



from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, Future
from typing import Sequence, Tuple, List, Optional, Callable, Any

from more_itertools import flatten

from resoto_plugin_aws.resource.base import ExecutorQueue, AwsRegion, GraphBuilder, GatherFutures
from resoto_plugin_aws.resource.ec2 import AwsEc2InstanceType
from test import account_collector, builder, aws_client, aws_config, no_feedback  # noqa: F401


def check_executor_work(
    work: List[Tuple[str, int]],
    workers: int,
    tasks_per_key: int,
    fail_on_first_exception: bool,
    check_in_progress: Optional[Callable[[set[int]], None]] = None,
) -> Tuple[List[int], Optional[Exception]]:
    work_done = []
    in_progress = set()

    def do_work(num: int) -> None:
        in_progress.add(num)
        if check_in_progress:
            check_in_progress(in_progress)
        in_progress.remove(num)
        if num % 100 == 5:
            raise Exception(f"Abort {num}")
        work_done.append(num)

    with ThreadPoolExecutor(max_workers=workers) as executor:
        queue = ExecutorQueue(
            executor, lambda _: tasks_per_key, "test", fail_on_first_exception_in_group=fail_on_first_exception
        )
        for key, idx in work:
            queue.submit_work(key, do_work, idx)

        try:
            queue.wait_for_submitted_work()
            return work_done, None
        except Exception as ex:
            return work_done, ex


def check_executor_queue(work: Sequence[int], fail_on_first_exception: bool) -> Tuple[List[int], Optional[Exception]]:
    return check_executor_work([("same_key", i) for i in work], 1, 1, fail_on_first_exception)


def test_parallel_work() -> None:
    def assert_one_per_key(in_progress: set[int]) -> None:
        assert len([a for a in in_progress if a < 100]) <= 2
        assert len([a for a in in_progress if a > 100]) <= 2

    # create work with 2 keys: w1(<100) and w2(>100)
    work = flatten([[("w1", i), ("w2", i + 100)] for i in range(10)])
    # test with 4 worker threads and max 2 tasks per key
    work_done, ex = check_executor_work(list(work), 4, 2, False, assert_one_per_key)
    assert ex is not None
    assert set(work_done) == {0, 1, 2, 3, 100, 101, 102, 103, 104, 4, 106, 6, 107, 7, 108, 8, 9, 109}


def test_success_queue() -> None:
    work_done, ex = check_executor_queue(range(5), True)
    assert work_done == [0, 1, 2, 3, 4]
    assert ex is None
    work_done, ex = check_executor_queue(range(5), False)
    assert work_done == [0, 1, 2, 3, 4]
    assert ex is None


def test_fail_on_first() -> None:
    work_done, ex = check_executor_queue(range(10), True)
    assert work_done == [0, 1, 2, 3, 4]
    assert ex is not None
    assert ex.args[0] == "Abort 5"


def test_fail_late() -> None:
    work_done, ex = check_executor_queue(range(10), False)
    assert work_done == [0, 1, 2, 3, 4, 6, 7, 8, 9]
    assert ex is not None
    assert ex.args[0] == "Abort 5"


def test_instance_type_handling(builder: GraphBuilder) -> None:
    region1 = AwsRegion(id="us-east-1")
    region2 = AwsRegion(id="us-east-2")
    it = AwsEc2InstanceType(id="t3.micro")
    builder.global_instance_types[it.safe_name] = it
    it1: AwsEc2InstanceType = builder.instance_type(region1, it.safe_name)  # type: ignore
    assert it1.region() == region1
    it2: AwsEc2InstanceType = builder.instance_type(region2, it.safe_name)  # type: ignore
    assert it2.region() == region2
    assert it1 is not it2
    assert it1.chksum != it2.chksum


def test_future_gatherer() -> None:
    def do_work(num: int) -> int:
        # sleep(0.1)
        return num

    groups = []

    def check_done(f: Future[Any]) -> None:
        groups.append(f)

    with ThreadPoolExecutor(max_workers=1) as executor:
        work = defaultdict(list)
        for i in range(100):
            work[i % 10].append(executor.submit(do_work, i))

        for group in work.values():
            GatherFutures.all(group).add_done_callback(check_done)
    assert len(groups) == 10



from datetime import datetime
from typing import ClassVar, Dict, Optional, List, Type, Union

from attrs import define, field
from resoto_plugin_aws.aws_client import AwsClient

from resoto_plugin_aws.resource.base import AwsResource, GraphBuilder, AwsApiSpec
from resoto_plugin_aws.resource.ec2 import AwsEc2VpcEndpoint
from resoto_plugin_aws.resource.iam import AwsIamRole
from resoto_plugin_aws.resource.route53 import AwsRoute53Zone

from resotolib.baseresources import EdgeType, ModelReference
from resotolib.graph import Graph
from resotolib.json import from_json
from resotolib.json_bender import Bender, S, Bend, bend
from resotolib.types import Json

service_name = "apigateway"


class ApiGatewayTaggable:
    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        if isinstance(self, AwsResource):
            if spec := self.api_spec:
                client.call(
                    aws_service=spec.service,
                    action="tag-resource",
                    result_name=None,
                    resourceArn=self.arn,
                    tags={key: value},
                )
                return True
            return False
        return False

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        if isinstance(self, AwsResource):
            if spec := self.api_spec:
                client.call(
                    aws_service=spec.service,
                    action="untag-resource",
                    result_name=None,
                    resourceArn=self.arn,
                    tagKeys=[key],
                )
                return True
            return False
        return False

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return [
            AwsApiSpec(service_name, "tag-resource", override_iam_permission="apigateway:PATCH"),
            AwsApiSpec(service_name, "tag-resource", override_iam_permission="apigateway:POST"),
            AwsApiSpec(service_name, "tag-resource", override_iam_permission="apigateway:PUT"),
            AwsApiSpec(service_name, "untag-resource", override_iam_permission="apigateway:DELETE"),
        ]


@define(eq=False, slots=False)
class AwsApiGatewayMethodResponse:
    kind: ClassVar[str] = "aws_api_gateway_method_response"
    mapping: ClassVar[Dict[str, Bender]] = {
        "status_code": S("statusCode"),
        "response_parameters": S("responseParameters"),
        "response_models": S("responseModels"),
    }
    status_code: Optional[str] = field(default=None)
    response_parameters: Optional[Dict[str, bool]] = field(default=None)
    response_models: Optional[Dict[str, str]] = field(default=None)


@define(eq=False, slots=False)
class AwsApiGatewayIntegrationResponse:
    kind: ClassVar[str] = "aws_api_gateway_integration_response"
    mapping: ClassVar[Dict[str, Bender]] = {
        "status_code": S("statusCode"),
        "selection_pattern": S("selectionPattern"),
        "response_parameters": S("responseParameters"),
        "response_templates": S("responseTemplates"),
        "content_handling": S("contentHandling"),
    }
    status_code: Optional[str] = field(default=None)
    selection_pattern: Optional[str] = field(default=None)
    response_parameters: Optional[Dict[str, str]] = field(default=None)
    response_templates: Optional[Dict[str, str]] = field(default=None)
    content_handling: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsApiGatewayIntegration:
    kind: ClassVar[str] = "aws_api_gateway_integration"
    mapping: ClassVar[Dict[str, Bender]] = {
        "integration_type": S("type"),
        "http_method": S("httpMethod"),
        "uri": S("uri"),
        "connection_type": S("connectionType"),
        "connection_id": S("connectionId"),
        "credentials": S("credentials"),
        "request_parameters": S("requestParameters"),
        "request_templates": S("requestTemplates"),
        "passthrough_behavior": S("passthroughBehavior"),
        "content_handling": S("contentHandling"),
        "timeout_in_millis": S("timeoutInMillis"),
        "cache_namespace": S("cacheNamespace"),
        "cache_key_parameters": S("cacheKeyParameters", default=[]),
        "integration_responses": S("integrationResponses"),
        "tls_config": S("tlsConfig", "insecureSkipVerification"),
    }
    integration_type: Optional[str] = field(default=None)
    http_method: Optional[str] = field(default=None)
    uri: Optional[str] = field(default=None)
    connection_type: Optional[str] = field(default=None)
    connection_id: Optional[str] = field(default=None)
    credentials: Optional[str] = field(default=None)
    request_parameters: Optional[Dict[str, str]] = field(default=None)
    request_templates: Optional[Dict[str, str]] = field(default=None)
    passthrough_behavior: Optional[str] = field(default=None)
    content_handling: Optional[str] = field(default=None)
    timeout_in_millis: Optional[int] = field(default=None)
    cache_namespace: Optional[str] = field(default=None)
    cache_key_parameters: List[str] = field(factory=list)
    integration_responses: Optional[Dict[str, AwsApiGatewayIntegrationResponse]] = field(default=None)
    tls_config: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class AwsApiGatewayMethod:
    kind: ClassVar[str] = "aws_api_gateway_method"
    mapping: ClassVar[Dict[str, Bender]] = {
        "http_method": S("httpMethod"),
        "authorization_type": S("authorizationType"),
        "authorizer_id": S("authorizerId"),
        "api_key_required": S("apiKeyRequired"),
        "request_validator_id": S("requestValidatorId"),
        "operation_name": S("operationName"),
        "request_parameters": S("requestParameters"),
        "request_models": S("requestModels"),
        "method_responses": S("methodResponses"),
        "method_integration": S("methodIntegration") >> Bend(AwsApiGatewayIntegration.mapping),
        "authorization_scopes": S("authorizationScopes", default=[]),
    }
    http_method: Optional[str] = field(default=None)
    authorization_type: Optional[str] = field(default=None)
    authorizer_id: Optional[str] = field(default=None)
    api_key_required: Optional[bool] = field(default=None)
    request_validator_id: Optional[str] = field(default=None)
    operation_name: Optional[str] = field(default=None)
    request_parameters: Optional[Dict[str, bool]] = field(default=None)
    request_models: Optional[Dict[str, str]] = field(default=None)
    method_responses: Optional[Dict[str, AwsApiGatewayMethodResponse]] = field(default=None)
    method_integration: Optional[AwsApiGatewayIntegration] = field(default=None)
    authorization_scopes: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsApiGatewayResource(AwsResource):
    # collection of resource resources happens in AwsApiGatewayRestApi.collect()
    kind: ClassVar[str] = "aws_api_gateway_resource"
    reference_kinds: ClassVar[ModelReference] = {"successors": {"default": ["aws_api_gateway_authorizer"]}}
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("id"),
        "resource_parent_id": S("parentId"),
        "resource_path_part": S("pathPart"),
        "resource_path": S("path"),
        "resource_methods": S("resourceMethods"),
    }
    resource_parent_id: Optional[str] = field(default=None)
    resource_path_part: Optional[str] = field(default=None)
    resource_path: Optional[str] = field(default=None)
    resource_methods: Optional[Dict[str, AwsApiGatewayMethod]] = field(default=None)
    api_link: Optional[str] = field(default=None)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if self.resource_methods:
            for method in self.resource_methods:
                builder.add_edge(
                    self,
                    edge_type=EdgeType.default,
                    clazz=AwsApiGatewayAuthorizer,
                    id=self.resource_methods[method].authorizer_id,
                )

    def delete_resource(self, client: AwsClient, graph: Graph) -> bool:
        client.call(
            aws_service=service_name,
            action="delete-resource",
            result_name=None,
            restApiId=self.api_link,
            resourceId=self.id,
        )
        return True

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return [AwsApiSpec(service_name, "delete-resource", override_iam_permission="apigateway:DELETE")]


@define(eq=False, slots=False)
class AwsApiGatewayAuthorizer(AwsResource):
    # collection of authorizer resources happens in AwsApiGatewayRestApi.collect()
    kind: ClassVar[str] = "aws_api_gateway_authorizer"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {"default": ["aws_lambda_function"]},
        "predecessors": {"default": ["aws_iam_role"], "delete": ["aws_lambda_function", "aws_iam_role"]},
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("id"),
        "name": S("name"),
        "authorizer_type": S("type"),
        "authorizer_provider_arns": S("providerARNs", default=[]),
        "authorizer_auth_type": S("authType"),
        "authorizer_uri": S("authorizerUri"),
        "authorizer_credentials": S("authorizerCredentials"),
        "authorizer_identity_source": S("identitySource"),
        "authorizer_identity_validation_expression": S("identityValidationExpression"),
        "authorizer_result_ttl_in_seconds": S("authorizerResultTtlInSeconds"),
    }
    authorizer_type: Optional[str] = field(default=None)
    authorizer_provider_arns: List[Optional[str]] = field(default=None)
    authorizer_auth_type: Optional[str] = field(default=None)
    authorizer_uri: Optional[str] = field(default=None)
    authorizer_credentials: Optional[str] = field(default=None)
    authorizer_identity_source: Optional[str] = field(default=None)
    authorizer_identity_validation_expression: Optional[str] = field(default=None)
    authorizer_result_ttl_in_seconds: Optional[int] = field(default=None)
    api_link: Optional[str] = field(default=None)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if self.authorizer_uri:
            lambda_name = self.authorizer_uri.split(":")[-1].removesuffix("/invocations")
            builder.dependant_node(
                self,
                kind="aws_lambda_function",
                name=lambda_name,
            )
        if self.authorizer_credentials:
            builder.dependant_node(
                self, reverse=True, delete_same_as_default=True, clazz=AwsIamRole, arn=self.authorizer_credentials
            )
        for user_pool in self.authorizer_provider_arns:
            builder.add_edge(self, edge_type=EdgeType.default, kind="aws_cognito_user_pool", arn=user_pool)

    def delete_resource(self, client: AwsClient, graph: Graph) -> bool:
        client.call(
            aws_service=service_name,
            action="delete-authorizer",
            result_name=None,
            restApiId=self.api_link,
            authorizerId=self.id,
        )
        return True

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return [AwsApiSpec(service_name, "delete-authorizer", override_iam_permission="apigateway:DELETE")]


@define(eq=False, slots=False)
class AwsApiGatewayCanarySetting:
    kind: ClassVar[str] = "aws_api_gateway_canary_setting"
    mapping: ClassVar[Dict[str, Bender]] = {
        "percent_traffic": S("percentTraffic"),
        "deployment_id": S("deploymentId"),
        "stage_variable_overrides": S("stageVariableOverrides"),
        "use_stage_cache": S("useStageCache"),
    }
    percent_traffic: Optional[int] = field(default=None)
    deployment_id: Optional[str] = field(default=None)
    stage_variable_overrides: Optional[Dict[str, str]] = field(default=None)
    use_stage_cache: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class AwsApiGatewayStage(ApiGatewayTaggable, AwsResource):
    # collection of stage resources happens in AwsApiGatewayRestApi.collect()
    kind: ClassVar[str] = "aws_api_gateway_stage"
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("syntheticId"),  # created by Resoto to avoid collision with duplicate stage names
        "name": S("stageName"),
        "tags": S("tags", default={}),
        "ctime": S("createdDate"),
        "mtime": S("lastUpdatedDate"),
        "description": S("description"),
        "stage_client_certificate_id": S("clientCertificateId"),
        "stage_cache_cluster_enabled": S("cacheClusterEnabled"),
        "stage_cache_cluster_size": S("cacheClusterSize"),
        "stage_cache_status": S("cacheClusterStatus"),
        "stage_method_settings": S("methodSettings"),
        "stage_variables": S("variables"),
        "stage_documentation_version": S("documentationVersion"),
        "stage_access_log_settings": S("accessLogSettings"),
        "stage_canary_settings": S("canarySettings") >> Bend(AwsApiGatewayCanarySetting.mapping),
        "stage_tracing_enabled": S("tracingEnabled"),
        "stage_web_acl_arn": S("webAclArn"),
    }
    description: Optional[str] = field(default=None)
    stage_client_certificate_id: Optional[str] = field(default=None)
    stage_cache_cluster_enabled: Optional[bool] = field(default=None)
    stage_cache_cluster_size: Optional[str] = field(default=None)
    stage_cache_status: Optional[str] = field(default=None)
    stage_method_settings: Optional[Dict[str, Dict[str, Union[bool, str, int]]]] = field(default=None)
    stage_variables: Optional[Dict[str, str]] = field(default=None)
    stage_documentation_version: Optional[str] = field(default=None)
    stage_access_log_settings: Optional[Dict[str, str]] = field(default=None)
    stage_canary_settings: Optional[AwsApiGatewayCanarySetting] = field(default=None)
    stage_tracing_enabled: Optional[bool] = field(default=None)
    stage_web_acl_arn: Optional[str] = field(default=None)
    api_link: Optional[str] = field(default=None)

    # TODO add edge to Web Acl when applicable (via stage_web_acl_arn)

    def delete_resource(self, client: AwsClient, graph: Graph) -> bool:
        client.call(
            aws_service=service_name,
            action="delete-stage",
            result_name=None,
            restApiId=self.api_link,
            stageName=self.name,
        )
        return True

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return super().called_mutator_apis() + [
            AwsApiSpec(service_name, "delete-stage", override_iam_permission="apigateway:DELETE")
        ]


@define(eq=False, slots=False)
class AwsApiGatewayDeployment(AwsResource):
    # collection of deployment resources happens in AwsApiGatewayRestApi.collect()
    kind: ClassVar[str] = "aws_api_gateway_deployment"
    # edge to aws_api_gateway_stage is established in AwsApiGatewayRestApi.collect()
    reference_kinds: ClassVar[ModelReference] = {"successors": {"default": ["aws_api_gateway_stage"]}}

    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("id"),
        "ctime": S("createdDate"),
        "description": S("description"),
        "deployment_api_summary": S("apiSummary"),
    }
    description: Optional[str] = field(default=None)
    deployment_api_summary: Optional[Dict[str, Dict[str, Dict[str, Union[str, bool]]]]] = field(default=None)
    api_link: Optional[str] = field(default=None)

    def delete_resource(self, client: AwsClient, graph: Graph) -> bool:
        client.call(
            aws_service=service_name,
            action="delete-deployment",
            result_name=None,
            restApiId=self.api_link,
            deploymentId=self.id,
        )
        return True

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return super().called_mutator_apis() + [
            AwsApiSpec(service_name, "delete-deployment", override_iam_permission="apigateway:DELETE")
        ]


@define(eq=False, slots=False)
class AwsApiGatewayEndpointConfiguration:
    kind: ClassVar[str] = "aws_api_gateway_endpoint_configuration"
    mapping: ClassVar[Dict[str, Bender]] = {
        "types": S("types", default=[]),
        "vpc_endpoint_ids": S("vpcEndpointIds", default=[]),
    }
    types: List[str] = field(factory=list)
    vpc_endpoint_ids: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsApiGatewayRestApi(ApiGatewayTaggable, AwsResource):
    kind: ClassVar[str] = "aws_api_gateway_rest_api"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec(
        service_name, "get-rest-apis", "items", override_iam_permission="apigateway:GET"
    )
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": [
                "aws_vpc_endpoint",
                "aws_api_gateway_deployment",
                "aws_api_gateway_authorizer",
                "aws_api_gateway_resource",
            ],
            "delete": ["aws_vpc_endpoint"],
        }
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("id"),
        "name": S("name"),
        "tags": S("tags", default={}),
        "ctime": S("createdDate"),
        "description": S("description"),
        "api_version": S("version"),
        "api_warnings": S("warnings", default=[]),
        "api_binary_media_types": S("binaryMediaTypes", default=[]),
        "api_minimum_compression_size": S("minimumCompressionSize"),
        "api_key_source": S("apiKeySource"),
        "api_endpoint_configuration": S("endpointConfiguration") >> Bend(AwsApiGatewayEndpointConfiguration.mapping),
        "api_policy": S("policy"),
        "api_disable_execute_api_endpoint": S("disableExecuteApiEndpoint"),
    }
    description: Optional[str] = field(default=None)
    api_version: Optional[str] = field(default=None)
    api_warnings: List[str] = field(factory=list)
    api_binary_media_types: List[str] = field(factory=list)
    api_minimum_compression_size: Optional[int] = field(default=None)
    api_key_source: Optional[str] = field(default=None)
    api_endpoint_configuration: Optional[AwsApiGatewayEndpointConfiguration] = field(default=None)
    api_policy: Optional[str] = field(default=None)
    api_disable_execute_api_endpoint: Optional[bool] = field(default=None)

    @classmethod
    def called_collect_apis(cls) -> List[AwsApiSpec]:
        return [
            cls.api_spec,
            AwsApiSpec(service_name, "get-deployments", override_iam_permission="apigateway:GET"),
            AwsApiSpec(service_name, "get-stages", override_iam_permission="apigateway:GET"),
            AwsApiSpec(service_name, "get-authorizers", override_iam_permission="apigateway:GET"),
            AwsApiSpec(service_name, "get-resources", override_iam_permission="apigateway:GET"),
        ]

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return super().called_mutator_apis() + [
            AwsApiSpec(service_name, "delete-rest-api", override_iam_permission="apigateway:DELETE")
        ]

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        for js in json:
            api_instance = cls.from_api(js)
            api_instance.set_arn(
                builder=builder,
                account="",
                resource=f"/restapis/{api_instance.id}",
            )
            builder.add_node(api_instance, js)
            for deployment in builder.client.list(service_name, "get-deployments", "items", restApiId=api_instance.id):
                deploy_instance = AwsApiGatewayDeployment.from_api(deployment)
                deploy_instance.set_arn(
                    builder=builder,
                    account="",
                    resource=f"/restapis/{api_instance.id}/deployments/{deploy_instance.id}",
                )
                deploy_instance.api_link = api_instance.id
                builder.add_node(deploy_instance, deployment)
                builder.add_edge(api_instance, EdgeType.default, node=deploy_instance)
                for stage in builder.client.list(
                    service_name, "get-stages", "item", restApiId=api_instance.id, deploymentId=deploy_instance.id
                ):
                    stage["syntheticId"] = f'{api_instance.id}_{stage["stageName"]}'  # create unique id
                    stage_instance = AwsApiGatewayStage.from_api(stage)
                    stage_instance.api_link = api_instance.id
                    builder.add_node(stage_instance, stage)
                    # reference kinds for this edge are maintained in AwsApiGatewayDeployment.reference_kinds
                    builder.add_edge(deploy_instance, EdgeType.default, node=stage_instance)
            for authorizer in builder.client.list(service_name, "get-authorizers", "items", restApiId=api_instance.id):
                auth_instance = AwsApiGatewayAuthorizer.from_api(authorizer)
                auth_instance.api_link = api_instance.id
                builder.add_node(auth_instance, authorizer)
                builder.add_edge(api_instance, EdgeType.default, node=auth_instance)
            for resource in builder.client.list(service_name, "get-resources", "items", restApiId=api_instance.id):
                resource_instance = AwsApiGatewayResource.from_api(resource)
                resource_instance.api_link = api_instance.id
                if resource_instance.resource_methods:
                    for method in resource_instance.resource_methods:
                        mapped = bend(AwsApiGatewayMethod.mapping, resource["resourceMethods"][method])
                        resource_instance.resource_methods[method] = from_json(mapped, AwsApiGatewayMethod)
                builder.add_node(resource_instance, resource)
                builder.add_edge(api_instance, EdgeType.default, node=resource_instance)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if self.api_endpoint_configuration:
            for endpoint in self.api_endpoint_configuration.vpc_endpoint_ids:
                builder.dependant_node(
                    self,
                    clazz=AwsEc2VpcEndpoint,
                    delete_same_as_default=True,
                    id=endpoint,
                )

    def delete_resource(self, client: AwsClient, graph: Graph) -> bool:
        client.call(
            aws_service=self.api_spec.service,
            action="delete-rest-api",
            result_name=None,
            restApiId=self.id,
        )
        return True


@define(eq=False, slots=False)
class AwsApiGatewayMutualTlsAuthentication:
    kind: ClassVar[str] = "aws_api_gateway_mutual_tls_authentication"
    mapping: ClassVar[Dict[str, Bender]] = {
        "truststore_uri": S("truststoreUri"),
        "truststore_version": S("truststoreVersion"),
        "truststore_warnings": S("truststoreWarnings", default=[]),
    }
    truststore_uri: Optional[str] = field(default=None)
    truststore_version: Optional[str] = field(default=None)
    truststore_warnings: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsApiGatewayDomainName(ApiGatewayTaggable, AwsResource):
    kind: ClassVar[str] = "aws_api_gateway_domain_name"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec(
        service_name, "get-domain-names", "items", override_iam_permission="apigateway:GET"
    )
    reference_kinds: ClassVar[ModelReference] = {
        "predecessors": {"delete": ["aws_route53_zone"]},
        "successors": {"default": ["aws_route53_zone", "aws_vpc_endpoint"], "delete": ["aws_vpc_endpoint"]},
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("domainName"),
        "tags": S("tags", default={}),
        "name": S("domainName"),
        "domain_certificate_name": S("certificateName"),
        "domain_certificate_arn": S("certificateArn"),
        "domain_certificate_upload_date": S("certificateUploadDate"),
        "domain_regional_domain_name": S("regionalDomainName"),
        "domain_regional_hosted_zone_id": S("regionalHostedZoneId"),
        "domain_regional_certificate_name": S("regionalCertificateName"),
        "domain_regional_certificate_arn": S("regionalCertificateArn"),
        "domain_distribution_domain_name": S("distributionDomainName"),
        "domain_distribution_hosted_zone_id": S("distributionHostedZoneId"),
        "domain_endpoint_configuration": S("endpointConfiguration") >> Bend(AwsApiGatewayEndpointConfiguration.mapping),
        "domain_domain_name_status": S("domainNameStatus"),
        "domain_domain_name_status_message": S("domainNameStatusMessage"),
        "domain_security_policy": S("securityPolicy"),
        "domain_mutual_tls_authentication": S("mutualTlsAuthentication")
        >> Bend(AwsApiGatewayMutualTlsAuthentication.mapping),
        "domain_ownership_verification_certificate_arn": S("ownershipVerificationCertificateArn"),
    }
    domain_certificate_name: Optional[str] = field(default=None)
    domain_certificate_arn: Optional[str] = field(default=None)
    domain_certificate_upload_date: Optional[datetime] = field(default=None)
    domain_regional_domain_name: Optional[str] = field(default=None)
    domain_regional_hosted_zone_id: Optional[str] = field(default=None)
    domain_regional_certificate_name: Optional[str] = field(default=None)
    domain_regional_certificate_arn: Optional[str] = field(default=None)
    domain_distribution_domain_name: Optional[str] = field(default=None)
    domain_distribution_hosted_zone_id: Optional[str] = field(default=None)
    domain_endpoint_configuration: Optional[AwsApiGatewayEndpointConfiguration] = field(default=None)
    domain_domain_name_status: Optional[str] = field(default=None)
    domain_domain_name_status_message: Optional[str] = field(default=None)
    domain_security_policy: Optional[str] = field(default=None)
    domain_mutual_tls_authentication: Optional[AwsApiGatewayMutualTlsAuthentication] = field(default=None)
    domain_ownership_verification_certificate_arn: Optional[str] = field(default=None)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if zid := self.domain_regional_hosted_zone_id:
            builder.dependant_node(self, clazz=AwsRoute53Zone, delete_same_as_default=True, id=zid)
        if configuration := self.domain_endpoint_configuration:
            for endpoint in configuration.vpc_endpoint_ids:
                builder.dependant_node(
                    self,
                    clazz=AwsEc2VpcEndpoint,
                    delete_same_as_default=True,
                    id=endpoint,
                )
        # TODO add edge to ACM Certificates when applicable


resources: List[Type[AwsResource]] = [
    AwsApiGatewayRestApi,
    AwsApiGatewayDeployment,
    AwsApiGatewayStage,
    AwsApiGatewayResource,
    AwsApiGatewayAuthorizer,
    AwsApiGatewayDomainName,
]



from resoto_plugin_aws.resource.elb import AwsElb
from resotolib.graph import Graph
from test.resources import round_trip_for
from typing import Any, cast
from types import SimpleNamespace
from resoto_plugin_aws.aws_client import AwsClient


def test_elbs() -> None:
    first, graph = round_trip_for(AwsElb, "public_ip_address")
    assert len(first.tags) == 2


def test_elb_deletion() -> None:
    elb, _ = round_trip_for(AwsElb, "public_ip_address")

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "delete-load-balancer"
        assert kwargs["LoadBalancerName"] == elb.name

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    elb.delete_resource(client, Graph())


def test_tagging() -> None:
    elb, _ = round_trip_for(AwsElb, "public_ip_address")

    def validate_update_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "add-tags"
        assert kwargs["LoadBalancerNames"] == [elb.name]
        assert kwargs["Tags"] == [{"Key": "foo", "Value": "bar"}]

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "remove-tags"
        assert kwargs["LoadBalancerNames"] == [elb.name]
        assert kwargs["Tags"] == [{"Key": "foo"}]

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    elb.update_resource_tag(client, "foo", "bar")

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    elb.delete_resource_tag(client, "foo")
